import sys
from jschema_to_python.python_file_generator import PythonFileGenerator
import jschema_to_python.utilities as util


class ClassGenerator(PythonFileGenerator):
    def __init__(self, class_schema, class_name, code_gen_hints, output_directory):
        super(ClassGenerator, self).__init__(output_directory)
        self.class_schema = class_schema
        self.required_property_names = class_schema.get("required")
        if self.required_property_names:
            self.required_property_names.sort()
        self.class_name = class_name
        self.code_gen_hints = code_gen_hints
        self.file_path = self.make_class_file_path()

    def __del__(self):
        sys.stdout = sys.__stdout__

    def generate(self):
        with open(self.file_path, "w") as sys.stdout:
            self.write_generation_comment()
            self.write_class_declaration()
            self.write_class_description()
            self.write_class_body()

    def make_class_file_path(self):
        class_module_name = util.class_name_to_private_module_name(self.class_name)
        return self.make_output_file_path(class_module_name + ".py")

    def write_class_declaration(self):
        print("import attr")
        print("")
        print("")  # The black formatter wants two blank lines here.
        print("@attr.s")
        print("class " + self.class_name + "(object):")

    def write_class_description(self):
        description = self.class_schema.get("description")
        if description:
            print('    """' + description + '"""')
            print("")  # The black formatter wants a blank line here.

    def write_class_body(self):
        property_schemas = self.class_schema["properties"]
        if not property_schemas:
            print("    pass")
            return

        schema_property_names = sorted(property_schemas.keys())

        # attrs requires that mandatory attributes be declared before optional
        # attributes.
        if self.required_property_names:
            for schema_property_name in self.required_property_names:
                attrib = self.make_attrib(schema_property_name)
                print(attrib)

        for schema_property_name in schema_property_names:
            if self.is_optional(schema_property_name):
                attrib = self.make_attrib(schema_property_name)
                print(attrib)

    def make_attrib(self, schema_property_name):
        python_property_name = self.make_python_property_name_from_schema_property_name(
            schema_property_name
        )
        attrib = "".join(["    ", python_property_name, " = attr.ib("])
        if self.is_optional(schema_property_name):
            property_schema = self.class_schema["properties"][schema_property_name]
            default_setter = self.make_default_setter(property_schema)
            attrib = "".join([attrib, default_setter, ", "])
        attrib = "".join([attrib, "metadata={\"schema_property_name\": \"", schema_property_name, "\"})"])
        return attrib

    def is_optional(self, schema_property_name):
        return (
            not self.required_property_names
            or schema_property_name not in self.required_property_names
        )

    def make_default_setter(self, property_schema):
        initializer = self.make_initializer(property_schema)
        return "default=" + str(initializer)

    def make_initializer(self, property_schema):
        default = property_schema.get("default")
        if default:
            type = property_schema.get("type")
            if type:
                if type == "string":
                    default = (
                        '"' + default + '"'
                    )  # The black formatter wants double-quotes.
                elif type == "array":
                    # It isn't safe to specify a mutable object as a default value,
                    # because all new instances share the same mutable object, and
                    # one of them might mutate it, affecting all future instances!
                    # attr.Factory creates a new value for each instance.
                    default="attr.Factory(lambda: " + str(default) + ")"
            elif property_schema.get("enum"):
                default = '"' + default + '"'
            return default

        return "None"

    def make_python_property_name_from_schema_property_name(self, schema_property_name):
        hint_key = self.class_name + "." + schema_property_name
        property_name_hint = self.get_hint(hint_key, "PropertyNameHint")
        if not property_name_hint:
            property_name = schema_property_name
        else:
            property_name = property_name_hint["arguments"]["pythonPropertyName"]
        return util.to_underscore_separated_name(property_name)

    def get_hint(self, hint_key, hint_kind):
        if not self.code_gen_hints or hint_key not in self.code_gen_hints:
            return None

        hint_array = self.code_gen_hints[hint_key]
        for hint in hint_array:
            if hint["kind"] == hint_kind:
                return hint

        return None



import argparse

from jschema_to_python.object_model_module_generator import ObjectModelModuleGenerator
from jschema_to_python import __version__


def main():
    parser = init_parser()
    args = parser.parse_args()

    display_args(args)

    generator = ObjectModelModuleGenerator(args)
    generator.generate()

    if args.verbose:
        print("Done.")


def init_parser():
    parser = argparse.ArgumentParser(
        description="Generate source code for a set of Python classes from a JSON schema."
    )
    parser.add_argument(
        "-s", "--schema-path", help="path to the JSON schema file", required=True
    )
    parser.add_argument(
        "-o",
        "--output-directory",
        help="directory in which the generated classes will be created",
        required=True,
    )
    parser.add_argument(
        "-m",
        "--module-name",
        help="name of the module containing the object model classes",
    )
    parser.add_argument(
        "-r",
        "--root-class-name",
        help="the name of the class at the root of the object model represented by the schema",
        required=True,
    )
    parser.add_argument(
        "-g",
        "--hints-file-path",
        help="path to a file containing hints that control code generation",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="overwrite the output directory if it exists",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase output verbosity (may be specified up to two times)",
    )
    return parser


def display_args(args):
    if args.verbose:
        print(
            __package__
            + ": JSON schema to Python object model generator, version "
            + __version__
        )
        print("Generating Python classes...")
    if args.verbose > 1:
        print("    from JSON schema " + args.schema_path)
        print("    to module " + args.module_name)
        print("    in directory " + args.output_directory)
        print("    with root class " + args.root_class_name)
        if args.hints_file_path:
            print("    with code generation hints from " + args.hints_file_path)


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft.  All Rights Reserved.
# Licensed under the MIT license. See LICENSE file in the project root for full license information.

import attr
import copy
import json


def to_json(obj):
    """''Serializes an instance of a generated class to JSON.

    :param obj: an instance of any class generated by jschema-to-python.

    Before serializing the instance to JSON, this function maps the Python
    property names to the corresponding JSON schema property names. It also
    removes any property that has the default value specified by attr.ib,
    making the resulting JSON smaller.
    """

    return json.dumps(obj, indent=2, default=_generated_class_serializer)


def _generated_class_serializer(obj):
    if hasattr(obj, "__dict__"):
        dict = getattr(obj, "__dict__")
        dict = copy.deepcopy(dict)
        _remove_properties_with_default_values(obj, dict)
        _change_python_property_names_to_schema_property_names(obj, dict)
        return dict
    else:
        return str(obj)


def _remove_properties_with_default_values(obj, dict):
    for field in attr.fields(obj.__class__):
        dict_entry = dict.get(field.name)
        if value_is_default(dict_entry, field.default) and field.name in dict:
            del dict[field.name]


def value_is_default(dict_entry, default_value):
    if type(default_value) == attr._make.Factory:
        value = default_value.factory()
    else:
        value = default_value
    return dict_entry == value


def _change_python_property_names_to_schema_property_names(obj, dict):
    for field in attr.fields(obj.__class__):
        schema_property_name = field.metadata.get("schema_property_name")
        if schema_property_name and schema_property_name != field.name and field.name in dict:
            dict[schema_property_name] = dict[field.name]
            del dict[field.name]



import os

import jschema_to_python.utilities as util
from jschema_to_python.init_file_generator import InitFileGenerator
from jschema_to_python.class_generator import ClassGenerator


class ObjectModelModuleGenerator:
    def __init__(self, args):
        self.output_directory = args.output_directory
        self.force = args.force
        self.module_name = args.module_name
        self.root_schema = self.read_schema(args.schema_path)
        self.code_gen_hints = self.read_code_gen_hints(args.hints_file_path)
        self.root_class_name = args.root_class_name

    def generate(self):
        util.create_directory(self.output_directory, self.force)
        self.generate_root_class()
        self.generate_definition_classes()
        self.generate_init_file()

    def generate_init_file(self):
        init_file_generator = InitFileGenerator(
            self.module_name,
            self.root_schema,
            self.root_class_name,
            self.output_directory,
        )
        init_file_generator.generate()

    def generate_root_class(self):
        class_generator = ClassGenerator(
            self.root_schema,
            self.root_class_name,
            self.code_gen_hints,
            self.output_directory,
        )
        class_generator.generate()

    def generate_definition_classes(self):
        definition_schemas = self.root_schema.get("definitions")
        if definition_schemas:
            for key in definition_schemas:
                self.generate_definition_class(key, definition_schemas[key])

    def generate_definition_class(self, definition_key, definition_schema):
        class_name = util.capitalize_first_letter(definition_key)
        class_generator = ClassGenerator(
            definition_schema, class_name, self.code_gen_hints, self.output_directory
        )
        class_generator.generate()

    def read_schema(self, schema_path):
        if not os.path.exists(schema_path):
            util.exit_with_error("schema file {} does not exist", schema_path)

        return util.unpickle_file(schema_path)

    def read_code_gen_hints(self, hints_file_path):
        if not hints_file_path:
            return None

        if not os.path.exists(hints_file_path):
            util.exit_with_error(
                "code generation hints file {} does not exist", hints_file_path
            )

        return util.unpickle_file(hints_file_path)
